import knexCleaner from 'knex-cleaner';
import {Resource} from '../database/models';
import knex from '../lib/knex';
import {
  getUserOwnedSpaces,
  getUserSpaces,
  getUserResources,
  getUserPermissions,
  getSpaceByName,
  getSpaceTeams,
  getSpaceOwners,
  getSpaceResources,
  getResource,
  getTeamAndMembers,
} from '../lib/queries';
import {sortByKey, sortByKeys} from '../lib/utils';

jest.setTimeout(30000);

beforeEach(async () => {
  await knexCleaner.clean(knex, {ignoreTables: ['knex_migrations']});
  await knex.migrate.latest();
  await knex.seed.run();
});

afterEach(async () => {
  await knex.destroy();
  await knex.initialize();
});

test('get owned spaces', async () => {
  const spaces1 = await getUserOwnedSpaces(1);
  spaces1.sort((a, b) => a.space_id - b.space_id);
  expect(spaces1).toEqual([
    {name: 'hsrn', fullname: 'nyu/hsrn', parent_space_id: 0, space_id: 1},
    {
      name: 'dashboard',
      fullname: 'nyu/hsrn/dashboard',
      parent_space_id: 1,
      space_id: 2,
    },
    {
      name: 'bob-class-project',
      fullname: 'nyu/hsrn/bob-class-project',
      parent_space_id: 1,
      space_id: 3,
    },
  ]);

  const spaces2 = await getUserOwnedSpaces(3);
  spaces2.sort((a, b) => a.space_id - b.space_id);
  expect(spaces2).toEqual([
    {
      name: 'dashboard',
      fullname: 'nyu/hsrn/dashboard',
      parent_space_id: 1,
      space_id: 2,
    },
  ]);
});

test('get all user spaces', async () => {
  const spaces1 = await getUserSpaces(1);
  spaces1.sort((a, b) => a.space.space_id - b.space.space_id);
  expect(spaces1).toEqual([
    {
      space: {name: 'nyu', fullname: 'nyu', parent_space_id: null, space_id: 0},
      team: {
        name: 'researchers',
        space_id: 0,
        team_id: 0,
      },
    },
    {
      space: {
        name: 'hsrn',
        fullname: 'nyu/hsrn',
        parent_space_id: 0,
        space_id: 1,
      },
      team: 'owners',
    },
    {
      space: {
        name: 'dashboard',
        fullname: 'nyu/hsrn/dashboard',
        parent_space_id: 1,
        space_id: 2,
      },
      team: 'owners',
    },
    {
      space: {
        name: 'bob-class-project',
        fullname: 'nyu/hsrn/bob-class-project',
        parent_space_id: 1,
        space_id: 3,
      },
      team: 'owners',
    },
  ]);

  const spaces2 = await getUserSpaces(3);
  spaces2.sort((a, b) => a.space.space_id - b.space.space_id);
  expect(spaces2).toEqual([
    {
      space: {name: 'nyu', fullname: 'nyu', parent_space_id: null, space_id: 0},
      team: {
        name: 'researchers',
        space_id: 0,
        team_id: 0,
      },
    },
    {
      space: {
        name: 'hsrn',
        fullname: 'nyu/hsrn',
        parent_space_id: 0,
        space_id: 1,
      },
      team: {
        name: 'leadership',
        space_id: 1,
        team_id: 2,
      },
    },
    {
      space: {
        name: 'dashboard',
        fullname: 'nyu/hsrn/dashboard',
        parent_space_id: 1,
        space_id: 2,
      },
      team: 'owners',
    },
  ]);

  const spaces3 = await getUserSpaces(4);
  spaces3.sort((a, b) => a.space.space_id - b.space.space_id);
  expect(spaces3).toEqual([
    {
      space: {name: 'nyu', fullname: 'nyu', parent_space_id: null, space_id: 0},
      team: {
        name: 'students',
        space_id: 0,
        team_id: 1,
      },
    },
    {
      space: {
        name: 'bob-class-project',
        fullname: 'nyu/hsrn/bob-class-project',
        parent_space_id: 1,
        space_id: 3,
      },
      team: 'owners',
    },
  ]);
});

test('get space by name', async () => {
  const space0 = await getSpaceByName(['nyu']);
  expect(space0).toEqual({
    space_id: 0,
    name: 'nyu',
    fullname: 'nyu',
    parent_space_id: null,
  });

  const space1 = await getSpaceByName(['nyu', 'hsrn']);
  expect(space1).toEqual({
    space_id: 1,
    name: 'hsrn',
    fullname: 'nyu/hsrn',
    parent_space_id: 0,
  });

  const space2 = await getSpaceByName(['nyu', 'hsrn', 'dashboard']);
  expect(space2).toEqual({
    space_id: 2,
    name: 'dashboard',
    fullname: 'nyu/hsrn/dashboard',
    parent_space_id: 1,
  });
});

test("get a space's teams", async () => {
  const teams0 = await getSpaceTeams(0);
  expect(teams0).toEqual([
    {
      team_id: 0,
      space_id: 0,
      name: 'researchers',
    },
    {
      team_id: 1,
      space_id: 0,
      name: 'students',
    },
  ]);

  const teams1 = await getSpaceTeams(1);
  expect(teams1).toEqual([
    {
      team_id: 2,
      space_id: 1,
      name: 'leadership',
    },
    {
      team_id: 3,
      space_id: 1,
      name: 'interns',
    },
  ]);
});

test("get a space's owners", async () => {
  const owners1 = await getSpaceOwners(1);
  expect(owners1).toEqual([
    {
      user: {
        user_id: 1,
        name: 'Rob Pahle',
        email_address: 'rp2585@nyu.edu',
        auth0_sub: 'oauth2|CILogon|http://cilogon.org/serverA/users/326236',
      },
      inherited: false,
    },
    {
      user: {
        user_id: 2,
        name: 'Jeremy Rowe',
        email_address: 'jr4121@nyu.edu',
        auth0_sub: 'oauth2|Other|jeremy',
      },
      inherited: false,
    },
  ]);

  const owners2 = await getSpaceOwners(2);
  sortByKey(owners2, e => e.user.user_id);
  expect(owners2).toEqual([
    {
      user: {
        user_id: 1,
        name: 'Rob Pahle',
        email_address: 'rp2585@nyu.edu',
        auth0_sub: 'oauth2|CILogon|http://cilogon.org/serverA/users/326236',
      },
      inherited: true,
    },
    {
      user: {
        user_id: 2,
        name: 'Jeremy Rowe',
        email_address: 'jr4121@nyu.edu',
        auth0_sub: 'oauth2|Other|jeremy',
      },
      inherited: true,
    },
    {
      user: {
        user_id: 3,
        name: 'Remi Rampin',
        email_address: 'remi.rampin@nyu.edu',
        auth0_sub: 'oauth2|CILogon|http://cilogon.org/serverA/users/9021171',
      },
      inherited: false,
    },
  ]);
});

test("get a space's resources", async () => {
  const res0 = await getSpaceResources(0);
  sortByKeys(res0, res => [res.resource_type, res.resource_id]);
  expect(res0).toEqual([
    {resource_type: '', resource_id: 'device'},
    {resource_type: '', resource_id: 'k8s'},
    {resource_type: '', resource_id: 'space'},
  ]);

  const res1 = await getSpaceResources(1);
  sortByKeys(res1, res => [res.resource_type, res.resource_id]);
  expect(res1).toEqual([
    {resource_type: '', resource_id: 'device'},
    {resource_type: '', resource_id: 'k8s'},
    {resource_type: '', resource_id: 'space'},
    {resource_type: 'k8s', resource_id: 'playground'},
    {resource_type: 'k8s', resource_id: 'zeek-elasticsearch'},
  ]);

  const res2 = await getSpaceResources(2);
  sortByKeys(res2, res => [res.resource_type, res.resource_id]);
  expect(res2).toEqual([
    {resource_type: '', resource_id: 'k8s'},
    {resource_type: '', resource_id: 'space'},
    {resource_type: 'k8s', resource_id: 'playground'},
  ]);
});

test("get a specific resource's details", async () => {
  const k8s_playground = await getResource({
    resource_type: 'k8s',
    resource_id: 'playground',
  });
  expect(k8s_playground).toEqual({
    resource_type: 'k8s',
    resource_id: 'playground',
    cpu_quota: '1500m',
    memory_quota: '60Gi',
  });

  const nonexistent = await getResource({
    resource_type: 'k8s',
    resource_id: 'doesntexist',
  });
  expect(nonexistent).toBeNull();
});

test('get a team and its members', async () => {
  // Get existing team
  const hsrn_leaders = await getTeamAndMembers(1, 'leadership');
  if (hsrn_leaders && hsrn_leaders.users) {
    sortByKey(hsrn_leaders.users, u => u.user_id);
  }
  expect(hsrn_leaders).toEqual({
    team: {
      space_id: 1,
      team_id: 2,
      name: 'leadership',
    },
    users: [
      {
        user_id: 1,
        name: 'Rob Pahle',
        email_address: 'rp2585@nyu.edu',
        auth0_sub: 'oauth2|CILogon|http://cilogon.org/serverA/users/326236',
      },
      {
        user_id: 2,
        name: 'Jeremy Rowe',
        email_address: 'jr4121@nyu.edu',
        auth0_sub: 'oauth2|Other|jeremy',
      },
      {
        user_id: 3,
        name: 'Remi Rampin',
        email_address: 'remi.rampin@nyu.edu',
        auth0_sub: 'oauth2|CILogon|http://cilogon.org/serverA/users/9021171',
      },
    ],
  });

  // Get empty team
  const hsrn_students = await getTeamAndMembers(1, 'interns');
  expect(hsrn_students).toEqual({
    team: {
      space_id: 1,
      team_id: 3,
      name: 'interns',
    },
    users: [],
  });

  // Get non-existent team
  const nonexistent = await getTeamAndMembers(0, 'leadership');
  expect(nonexistent).toBeNull();
});

test('user access to resources', async () => {
  // Deterministic ordering of results, the database will return in any order
  function doSort(
    resources: Array<{
      resource: Resource;
      space_id: number;
      permissions: string[];
    }>
  ) {
    for (const entry of resources) {
      entry.permissions.sort();
    }
    sortByKeys(resources, e => [
      e.resource.resource_type,
      e.resource.resource_id,
      e.space_id,
      e.permissions.join(','),
    ]);
  }

  const resources1 = await getUserResources(1);
  doSort(resources1);
  expect(resources1).toEqual([
    // User 1 owns space 1 'nyu/hsrn'
    // User 1 is in space 0 'nyu' team 0 'researchers'
    // User 1 is in space 1 'nyu/hsrn' team 2 'leadership'
    {
      resource: {resource_type: '', resource_id: 'device'},
      space_id: 0,
      permissions: ['register'],
    },
    {
      resource: {resource_type: '', resource_id: 'device'},
      space_id: 1,
      permissions: ['register'],
    },
    {
      resource: {resource_type: '', resource_id: 'k8s'},
      space_id: 0,
      permissions: ['createNamespace'],
    },
    {
      resource: {resource_type: '', resource_id: 'k8s'},
      space_id: 1,
      permissions: ['createNamespace'],
    },
    {
      resource: {resource_type: '', resource_id: 'k8s'},
      space_id: 2,
      permissions: ['createNamespace'],
    },
    {
      resource: {resource_type: '', resource_id: 'space'},
      space_id: 0,
      permissions: ['create'],
    },
    {
      resource: {resource_type: '', resource_id: 'space'},
      space_id: 1,
      permissions: ['create'],
    },
    {
      resource: {resource_type: '', resource_id: 'space'},
      space_id: 2,
      permissions: ['create'],
    },
    {
      resource: {resource_type: 'k8s', resource_id: 'playground'},
      space_id: 1,
      permissions: ['read', 'write'],
    },
    {
      resource: {resource_type: 'k8s', resource_id: 'playground'},
      space_id: 2,
      permissions: ['read', 'write'],
    },
    {
      resource: {resource_type: 'k8s', resource_id: 'playground'},
      space_id: 3,
      permissions: ['read'],
    },
    {
      resource: {resource_type: 'k8s', resource_id: 'zeek-elasticsearch'},
      space_id: 1,
      permissions: ['read', 'write'],
    },
  ]);
  expect(
    await getUserPermissions(1, {
      resource_type: 'k8s',
      resource_id: 'playground',
    })
  ).toEqual({read: true, write: true});
  expect(
    await getUserPermissions(
      1,
      {
        resource_type: 'k8s',
        resource_id: 'playground',
      },
      0
    )
  ).toEqual({});
  expect(
    await getUserPermissions(
      1,
      {
        resource_type: 'k8s',
        resource_id: 'playground',
      },
      1
    )
  ).toEqual({read: true, write: true});

  const resources3 = await getUserResources(3);
  doSort(resources3);
  expect(resources3).toEqual([
    // User 3 owns space 2 'nyu/hsrn/dashboard'
    // User 3 is in space 0 'nyu' team 0 'researchers'
    // User 3 is in space 1 'nyu/hsrn' team 2 'leadership'
    {
      resource: {resource_type: '', resource_id: 'device'},
      space_id: 0,
      permissions: ['register'],
    },
    {
      resource: {resource_type: '', resource_id: 'device'},
      space_id: 1,
      permissions: ['register'],
    },
    {
      resource: {resource_type: '', resource_id: 'k8s'},
      space_id: 0,
      permissions: ['createNamespace'],
    },
    {
      resource: {resource_type: '', resource_id: 'k8s'},
      space_id: 1,
      permissions: ['createNamespace'],
    },
    {
      resource: {resource_type: '', resource_id: 'k8s'},
      space_id: 2,
      permissions: ['createNamespace'],
    },
    {
      resource: {resource_type: '', resource_id: 'space'},
      space_id: 0,
      permissions: ['create'],
    },
    {
      resource: {resource_type: '', resource_id: 'space'},
      space_id: 1,
      permissions: ['create'],
    },
    {
      resource: {resource_type: '', resource_id: 'space'},
      space_id: 2,
      permissions: ['create'],
    },
    {
      resource: {resource_type: 'k8s', resource_id: 'playground'},
      space_id: 1,
      permissions: ['read', 'write'],
    },
    {
      resource: {resource_type: 'k8s', resource_id: 'playground'},
      space_id: 2,
      permissions: ['read', 'write'],
    },
    {
      resource: {resource_type: 'k8s', resource_id: 'zeek-elasticsearch'},
      space_id: 1,
      permissions: ['read', 'write'],
    },
  ]);
  expect(
    await getUserPermissions(3, {
      resource_type: 'k8s',
      resource_id: 'playground',
    })
  ).toEqual({read: true, write: true});
  expect(
    await getUserPermissions(3, {
      resource_type: 'k8s',
      resource_id: 'zeek-elasticsearch',
    })
  ).toEqual({read: true, write: true});

  const resources4 = await getUserResources(4);
  doSort(resources4);
  expect(resources4).toEqual([
    // User 4 is in space 0 'nyu' team 1 'students'
    // User 4 owns space 3 'nyu/hsrn/bob-class-project'
    {
      resource: {resource_type: '', resource_id: 'device'},
      space_id: 0,
      permissions: ['register'],
    },
    {
      resource: {resource_type: '', resource_id: 'space'},
      space_id: 0,
      permissions: ['create'],
    },
    {
      resource: {resource_type: 'k8s', resource_id: 'playground'},
      space_id: 3,
      permissions: ['read'],
    },
  ]);
  expect(
    await getUserPermissions(4, {
      resource_type: 'k8s',
      resource_id: 'playground',
    })
  ).toEqual({read: true});
});
