import {Knex} from 'knex';

export async function up(knex: Knex): Promise<void> {
  await knex.schema.createTable('users', table => {
    table.increments('user_id').primary();
    table.string('name', 255).notNullable();
    table.string('email_address', 255).nullable();
    table.string('auth0_sub', 255).nullable();
  });

  await knex.schema.createTable('resources', table => {
    table.string('resource_type', 255).notNullable();
    table.string('resource_id', 255).notNullable();
    table.primary(['resource_type', 'resource_id']);
  });

  await knex.schema.createTable('spaces', table => {
    table.increments('space_id').primary();
    table.string('name', 255).notNullable().checkNotIn(['']);
    table
      .integer('parent_space_id')
      .nullable()
      .references('space_id')
      .inTable('spaces');
  });

  await knex.schema.createTable('space_owners', table => {
    table
      .integer('space_id')
      .notNullable()
      .references('space_id')
      .inTable('spaces');
    table
      .integer('user_id')
      .notNullable()
      .references('user_id')
      .inTable('users');
    table.primary(['space_id', 'user_id']);
  });

  await knex.schema.createTable('teams', table => {
    table.increments('team_id').primary();
    table
      .integer('space_id')
      .notNullable()
      .references('space_id')
      .inTable('spaces');
    table.string('name', 255).notNullable().checkNotIn(['']);
  });

  await knex.schema.createTable('team_users', table => {
    table
      .integer('team_id')
      .notNullable()
      .references('team_id')
      .inTable('teams');
    table
      .integer('user_id')
      .notNullable()
      .references('user_id')
      .inTable('users');
    table.primary(['team_id', 'user_id']);
  });

  await knex.schema.createTable('space_resources', table => {
    table
      .integer('space_id')
      .notNullable()
      .references('space_id')
      .inTable('spaces');
    table.string('resource_type', 127).notNullable();
    table.string('resource_id', 255).notNullable();
    table.primary(['space_id', 'resource_type', 'resource_id']);
    table
      .foreign(['resource_type', 'resource_id'])
      .references(['resource_type', 'resource_id'])
      .inTable('resources');
  });

  await knex.schema.createTable('resource_permissions', table => {
    table.increments('id').primary();
    table
      .integer('space_id')
      .notNullable()
      .references('space_id')
      .inTable('spaces');
    table.string('resource_type', 127).notNullable();
    table.string('resource_id', 255).notNullable();
    table.integer('team_id').nullable().references('team_id').inTable('teams');
    table.integer('user_id').nullable().references('user_id').inTable('users');
    table.string('permission', 255).notNullable();
    table.boolean('apply_here').notNullable();
    table.boolean('inherit').notNullable();
    table
      .foreign(['resource_type', 'resource_id'])
      .references(['resource_type', 'resource_id'])
      .inTable('resources');
    table.unique([
      'space_id',
      'resource_id',
      'team_id',
      'user_id',
      'permission',
    ]);
  });

  await knex.schema.createTable('k8s_resources', table => {
    table.string('resource_type', 255).notNullable(); // Always 'k8s'
    table.string('resource_id', 255).notNullable();
    table.string('cpu_quota', 16).nullable();
    table.string('memory_quota', 16).nullable();
    table.primary(['resource_type', 'resource_id']);
    table
      .foreign(['resource_type', 'resource_id'])
      .references(['resource_type', 'resource_id'])
      .inTable('resources');
    table.check('resource_type = ?', ['k8s']);
  });

  await knex.schema.createTable('device_resources', table => {
    table.string('resource_type', 255).notNullable(); // Always 'device'
    table.string('resource_id', 255).notNullable();
    table.string('mac_address', 18).nullable();
    table.primary(['resource_type', 'resource_id']);
    table
      .foreign(['resource_type', 'resource_id'])
      .references(['resource_type', 'resource_id'])
      .inTable('resources');
    table.check('resource_type = ?', ['device']);
  });
}

export async function down(knex: Knex): Promise<void> {
  await knex.schema.dropTable('device_resources');
  await knex.schema.dropTable('k8s_resources');
  await knex.schema.dropTable('resource_permissions');
  await knex.schema.dropTable('space_resources');
  await knex.schema.dropTable('team_users');
  await knex.schema.dropTable('teams');
  await knex.schema.dropTable('space_owners');
  await knex.schema.dropTable('spaces');
  await knex.schema.dropTable('resources');
  await knex.schema.dropTable('users');
}
