import {GetServerSideProps, GetServerSidePropsResult} from 'next';
import {Counter, Histogram, collectDefaultMetrics} from 'prom-client';

interface GlobalMetrics {
  requestTotal: Counter;
  requestDuration: Histogram;
  requestErrors: Counter;
}

export function setupMetrics(): GlobalMetrics {
  interface GlobalData {
    prometheus_metrics?: GlobalMetrics;
  }

  if (!(global as unknown as GlobalData).prometheus_metrics) {
    (global as unknown as GlobalData).prometheus_metrics = {
      requestTotal: new Counter({
        name: 'request_total',
        help: 'Count of requests',
        labelNames: ['name'],
      }),
      requestDuration: new Histogram({
        name: 'request_seconds',
        help: 'Duration of requests',
        labelNames: ['name'],
      }),
      requestErrors: new Counter({
        name: 'request_errors_total',
        help: 'Count of requests raising an error',
        labelNames: ['name'],
      }),
    };
    collectDefaultMetrics();
  }
  return (global as unknown as GlobalData).prometheus_metrics!;
}

const {requestTotal, requestDuration, requestErrors} = setupMetrics();

// Middleware for getServerSideProps() that records metrics
export function reportMetrics<P>(
  name: string,
  func: GetServerSideProps<P>
): GetServerSideProps<P> {
  // Initialize counter
  requestTotal.labels({name}).inc(0);
  requestErrors.labels({name}).inc(0);

  // Wrapper function
  const wrappedFunc: GetServerSideProps<P> = async (
    context
  ): Promise<GetServerSidePropsResult<P>> => {
    // Increment counter
    requestTotal.labels({name}).inc();
    // Start timing
    const endTimer = requestDuration.labels({name}).startTimer();
    // Call wrapped function
    try {
      return await func(context);
    } catch (e) {
      requestErrors.labels({name}).inc();
      throw e;
    } finally {
      // End timing
      endTimer();
    }
  };
  return wrappedFunc;
}
