import knex from './knex';
import {Knex} from 'knex';
import {
  Resource,
  Space,
  Team,
  TeamOrOwners,
  User,
  ConcreteResource,
  K8sResource,
  DeviceResource,
} from '../database/models';

function ownedSpaceIds(userId: number): Knex.QueryBuilder<{}, unknown[]> {
  return knex.withRecursive('owned_spaceids', ['space_id'], qr => {
    qr.select('space_id')
      .from('space_owners')
      .where('space_owners.user_id', userId)
      .union(qu => {
        qu.select('spaces.space_id')
          .from('spaces')
          .join(
            'owned_spaceids',
            'owned_spaceids.space_id',
            'spaces.parent_space_id'
          );
      });
  });
}

function getSpaceFullName(): Knex.QueryBuilder<{}, unknown[]> {
  let concatenation;
  const dialect = knex.client.dialect;
  if (dialect === 'sqlite3') {
    concatenation = knex.raw('group_concat(name, ?)', ['/']);
  } else {
    concatenation = knex.raw('string_agg(name, ?)', ['/']);
  }
  return knex
    .withRecursive('parent_spaces', ['space_id', 'name', 'counter'], qr => {
      qr.select(
        'spaceFullName.parent_space_id',
        'spaceFullName.name',
        knex.raw('0')
      )
        .from({spaceFullName: 'spaces'})
        .where('spaceFullName.space_id', knex.ref('spaces.space_id'))
        .unionAll(qu => {
          qu.select(
            'spaceFullName2.parent_space_id',
            'spaceFullName2.name',
            knex.raw('parent_spaces.counter + 1')
          )
            .from({spaceFullName2: 'spaces'})
            .join(
              'parent_spaces',
              'parent_spaces.space_id',
              'spaceFullName2.space_id'
            );
        });
    })
    .select(concatenation)
    .from(
      knex
        .select('name')
        .from('parent_spaces')
        .orderBy('counter', 'desc')
        .as('spaceFullName3')
    );
}

// Get all spaces a user owns
export async function getUserOwnedSpaces(userId: number): Promise<Space[]> {
  const spaces = await ownedSpaceIds(userId)
    .select(
      'spaces.space_id',
      'spaces.name',
      'spaces.parent_space_id',
      getSpaceFullName().as('fullname')
    )
    .from('spaces')
    .join('owned_spaceids', 'owned_spaceids.space_id', 'spaces.space_id');
  return spaces;
}

// Get all spaces that a user has access to, with team name
export async function getUserSpaces(
  userId: number
): Promise<Array<{space: Space; team: TeamOrOwners}>> {
  const results = await ownedSpaceIds(userId)
    // Select the spaces owned
    .select(
      'spaces.space_id',
      'spaces.name',
      'spaces.parent_space_id',
      knex.raw('? AS team_id', [-1]),
      knex.raw('? AS team_name', ['']),
      getSpaceFullName().as('fullname')
    )
    .from('spaces')
    .join('owned_spaceids', 'owned_spaceids.space_id', 'spaces.space_id')
    // Union the teams
    .union(qu => {
      qu.select(
        'spaces.space_id',
        'spaces.name',
        'spaces.parent_space_id',
        'teams.team_id',
        knex.raw('teams.name AS team_name'),
        getSpaceFullName().as('fullname')
      )
        .from('spaces')
        .join('teams', 'teams.space_id', 'spaces.space_id')
        .join('team_users', 'team_users.team_id', 'teams.team_id')
        .where('team_users.user_id', userId)
        .whereNotIn(
          'teams.space_id',
          knex.select('space_id').from('owned_spaceids')
        );
    });
  return results.map(r => {
    const o: {space: Space; team: TeamOrOwners} = {
      space: {
        space_id: r.space_id,
        name: r.name,
        fullname: r.fullname,
        parent_space_id: r.parent_space_id,
      },
      team: 'owners',
    };
    if (r.team_id !== -1) {
      o.team = {
        team_id: r.team_id,
        name: r.team_name,
        space_id: r.space_id,
      };
    }
    return o;
  });
}

// Get the resources that a user has access to
export async function getUserResources(
  userId: number,
  resource_type?: string
): Promise<
  Array<{resource: Resource; space_id: number; permissions: string[]}>
> {
  // User has all permissions given to anyone in spaces user owns
  let permissionsFromOwnedSpaces = ownedSpaceIds(userId)
    .select(
      'resource_permissions.space_id',
      'resource_permissions.resource_type',
      'resource_permissions.resource_id',
      'resource_permissions.permission'
    )
    .from('resource_permissions')
    .join(
      'owned_spaceids',
      'owned_spaceids.space_id',
      'resource_permissions.space_id'
    )
    .where('resource_permissions.apply_here', true);
  if (resource_type !== undefined) {
    permissionsFromOwnedSpaces = permissionsFromOwnedSpaces.where(
      'resource_permissions.resource_type',
      resource_type
    );
  }

  // User has all permissions given to their teams
  let permissionsFromTeams = knex
    .select(
      'resource_permissions.space_id',
      'resource_permissions.resource_type',
      'resource_permissions.resource_id',
      'resource_permissions.permission'
    )
    .from('team_users')
    .join(
      'resource_permissions',
      'resource_permissions.team_id',
      'team_users.team_id'
    )
    .where('team_users.user_id', userId);
  if (resource_type !== undefined) {
    permissionsFromTeams = permissionsFromTeams.where(
      'resource_permissions.resource_type',
      resource_type
    );
  }

  // User has all permissions given directly to them
  let permissionsDirect = knex
    .select(
      'resource_permissions.space_id',
      'resource_permissions.resource_type',
      'resource_permissions.resource_id',
      'resource_permissions.permission'
    )
    .from('resource_permissions')
    .where('resource_permissions.user_id', userId);
  if (resource_type !== undefined) {
    permissionsDirect = permissionsDirect.where(
      'resource_permissions.resource_type',
      resource_type
    );
  }

  // Union and run
  const rows = await permissionsFromOwnedSpaces
    .union(permissionsFromTeams)
    .union(permissionsDirect);

  // Build result
  const resourceMap: {[id: string]: string[]} = {};
  for (const row of rows) {
    const id = row.space_id + ':' + row.resource_type + ':' + row.resource_id;
    if (resourceMap[id] === undefined) {
      resourceMap[id] = [row.permission];
    } else {
      resourceMap[id].push(row.permission);
    }
  }
  const result = [];
  for (const id of Object.getOwnPropertyNames(resourceMap)) {
    const sep1 = id.indexOf(':');
    const sep2 = id.indexOf(':', sep1 + 1);
    const space_id = parseInt(id.substring(0, sep1), 10);
    result.push({
      resource: {
        resource_type: id.substring(sep1 + 1, sep2),
        resource_id: id.substring(sep2 + 1),
      },
      space_id,
      permissions: resourceMap[id],
    });
  }
  return result;
}

// Get a user's permissions on a resource
export async function getUserPermissions(
  userId: number,
  resource: Resource,
  spaceId?: number
): Promise<{[permission: string]: true}> {
  // User has all permissions given to anyone in spaces user owns
  let permissionsFromOwnedSpaces = ownedSpaceIds(userId)
    .select('resource_permissions.permission')
    .from('resource_permissions')
    .join(
      'owned_spaceids',
      'owned_spaceids.space_id',
      'resource_permissions.space_id'
    )
    .where('resource_permissions.apply_here', true)
    .where('resource_permissions.resource_type', resource.resource_type)
    .where('resource_permissions.resource_id', resource.resource_id);
  if (spaceId !== undefined) {
    permissionsFromOwnedSpaces = permissionsFromOwnedSpaces.where(
      'resource_permissions.space_id',
      spaceId
    );
  }

  // User has all permissions given to their teams
  let permissionsFromTeams = knex
    .select('resource_permissions.permission')
    .from('team_users')
    .join(
      'resource_permissions',
      'resource_permissions.team_id',
      'team_users.team_id'
    )
    .where('team_users.user_id', userId)
    .where('resource_permissions.resource_type', resource.resource_type)
    .where('resource_permissions.resource_id', resource.resource_id);
  if (spaceId !== undefined) {
    permissionsFromTeams = permissionsFromTeams.where(
      'resource_permissions.space_id',
      spaceId
    );
  }

  // User has all permissions given directly to them
  let permissionsDirect = knex
    .select('resource_permissions.permission')
    .from('resource_permissions')
    .where('resource_permissions.user_id', userId)
    .where('resource_permissions.resource_type', resource.resource_type)
    .where('resource_permissions.resource_id', resource.resource_id);
  if (spaceId !== undefined) {
    permissionsDirect = permissionsDirect.where(
      'resource_permissions.space_id',
      spaceId
    );
  }

  // Union and run
  const rows = await permissionsFromOwnedSpaces
    .union(permissionsFromTeams)
    .union(permissionsDirect);
  const result: {[permission: string]: true} = {};
  for (const row of rows) {
    result[row.permission] = true;
  }
  return result;
}

// Get a space from its full name
export async function getSpaceByName(
  nameComponents: string[]
): Promise<Space | null> {
  if (nameComponents.length === 0) {
    return null;
  }
  if (nameComponents.length === 1) {
    const result = await knex<Space>('spaces')
      .where('name', nameComponents[0])
      .whereNull('parent_space_id')
      .first();
    if (result) {
      result.fullname = nameComponents.join('/');
      return result;
    } else {
      return null;
    }
  }
  let query = knex
    .select('space_id')
    .from('spaces')
    .where('name', nameComponents[0])
    .whereNull('parent_space_id');
  for (let i = 1; i < nameComponents.length - 1; ++i) {
    query = knex
      .select('space_id')
      .from('spaces')
      .where('name', nameComponents[i])
      .where('parent_space_id', query);
  }
  const result = knex<Space>('spaces')
    .where('name', nameComponents[nameComponents.length - 1])
    .where('parent_space_id', query)
    .first();
  const row = await result;
  if (row) {
    row.fullname = nameComponents.join('/');
    return row;
  } else {
    return null;
  }
}

// Get sub-spaces of a space
export async function getSubSpaces(spaceId: number): Promise<Space[]> {
  return await knex<Space>('spaces').where('parent_space_id', spaceId);
}

// Get the teams that are directly in a space
export async function getSpaceTeams(spaceId: number): Promise<Team[]> {
  return await knex<Team>('teams').where('space_id', spaceId);
}

// Get the owners of a space
export async function getSpaceOwners(
  spaceId: number
): Promise<Array<{user: User; inherited: boolean}>> {
  const rows = await knex
    .withRecursive('parent_spaces', ['space_id'], qr => {
      qr.select('parent_space_id')
        .from('spaces')
        .where('spaces.space_id', spaceId)
        .union(qu => {
          qu.select('spaces.parent_space_id')
            .from('spaces')
            .join(
              'parent_spaces',
              'parent_spaces.space_id',
              'spaces.parent_space_id'
            )
            .whereNotNull('spaces.parent_space_id');
        });
    })
    .select(
      'users.user_id',
      'users.name',
      'users.email_address',
      'users.auth0_sub',
      knex.raw('? as inherited', [0])
    )
    .from('users')
    .join('space_owners', 'space_owners.user_id', 'users.user_id')
    .where('space_id', spaceId)
    .unionAll(qu => {
      qu.select(
        'users.user_id',
        'users.name',
        'users.email_address',
        'users.auth0_sub',
        knex.raw('? as inherited', [1])
      )
        .from('users')
        .join('space_owners', 'space_owners.user_id', 'users.user_id')
        .join(
          'parent_spaces',
          'space_owners.space_id',
          'parent_spaces.space_id'
        );
    });
  return rows.map(r => ({
    user: {
      user_id: r.user_id,
      name: r.name,
      email_address: r.email_address,
      auth0_sub: r.auth0_sub,
    },
    inherited: r.inherited === 1 || r.inherited === true || r.inherited === '1',
  }));
}

// Get the resources in a space (not in sub-spaces)
export async function getSpaceResources(spaceId: number): Promise<Resource[]> {
  return await knex
    .select('space_resources.resource_type', 'space_resources.resource_id')
    .from('space_resources')
    .where('space_resources.space_id', spaceId);
}

// Get the given resource
export async function getResource({
  resource_type,
  resource_id,
}: Resource): Promise<ConcreteResource | null> {
  if (resource_type === 'k8s') {
    return (
      (await knex<K8sResource>('k8s_resources')
        .where('resource_type', resource_type)
        .where('resource_id', resource_id)
        .first()) || null
    );
  } else if (resource_type === 'device') {
    return (
      (await knex<DeviceResource>('device_resources')
        .where('resource_type', resource_type)
        .where('resource_id', resource_id)
        .first()) || null
    );
  } else {
    console.error('Unknown resource type ', resource_type);
    return null;
  }
}

// Get a team and its members
export async function getTeamAndMembers(
  spaceId: number,
  name: string
): Promise<{team: Team; users: User[]} | null> {
  const rows = await knex
    .select(
      'teams.team_id',
      'users.user_id',
      'users.name',
      'users.email_address',
      'users.auth0_sub'
    )
    .from('teams')
    .leftOuterJoin('team_users', 'teams.team_id', 'team_users.team_id')
    .leftOuterJoin('users', 'team_users.user_id', 'users.user_id')
    .where('teams.space_id', spaceId)
    .where('teams.name', name);
  if (rows.length === 0) {
    return null;
  }
  const team = {team_id: rows[0].team_id, space_id: spaceId, name};
  if (rows.length === 1 && rows[0].user_id === null) {
    // No users in this team, outer join just returned team information
    return {
      team,
      users: [],
    };
  } else {
    // We have users
    const users = rows.map(r => ({
      user_id: r.user_id,
      name: r.name,
      email_address: r.email_address,
      auth0_sub: r.auth0_sub,
    }));
    return {team, users};
  }
}
