import {GetServerSidePropsContext, GetServerSidePropsResult} from 'next';
import {getSession, withPageAuthRequired} from '@auth0/nextjs-auth0';
import {User} from '../database/models';
import {getOrCreateUser} from '../lib/get_user';
import {OidcApp, OidcUser, kubeconfig} from '../lib/kubeconfig';
import {reportMetrics} from '../lib/metrics';
import {Layout} from '../components/layout';

interface OidcInfo {
  oidcApp: OidcApp;
  oidcUser: OidcUser;
}

interface KubeConfigProps {
  user: User | null;
  oidcInfo: OidcInfo | null;
}

export default function KubeConfig({user, oidcInfo}: KubeConfigProps) {
  let configString;
  if (oidcInfo) {
    const {oidcApp, oidcUser} = oidcInfo;
    configString = kubeconfig(oidcApp, oidcUser);
  }
  return (
    <Layout title="Kubernetes configuration file" user={user}>
      <h1>Kubernetes configuration file</h1>
      <p>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed venenatis
        id eros vel sagittis. Mauris ornare ligula id diam euismod egestas.
        Donec mollis nibh sed orci tempus dapibus. Nullam non enim vitae ante
        efficitur rutrum vel sit amet quam. Duis quis quam sed odio convallis
        hendrerit. Etiam id fermentum metus, a commodo dui.
      </p>
      <pre>{configString}</pre>
    </Layout>
  );
}

export const getServerSideProps = reportMetrics(
  'kubeconfig',
  withPageAuthRequired({
    async getServerSideProps(
      ctx: GetServerSidePropsContext
    ): Promise<GetServerSidePropsResult<KubeConfigProps>> {
      const session = getSession(ctx.req, ctx.res);
      let oidcInfo: OidcInfo | null = null;
      let user = null;
      if (session) {
        user = await getOrCreateUser(session.user);
        if (session.refreshToken && session.accessToken && session.idToken) {
          oidcInfo = {
            oidcApp: {
              clientId: process.env.AUTH0_CLIENT_ID!,
              clientSecret: process.env.AUTH0_CLIENT_SECRET!,
              issuerUrl: process.env.AUTH0_ISSUER_BASE_URL!,
            },
            oidcUser: {
              userId: session.user.sub,
              refreshToken: session.refreshToken,
              accessToken: session.accessToken,
              idToken: session.idToken,
            },
          };
        } else {
          console.warn("User session doesn't have all required field");
        }
      }
      return {
        props: {
          user,
          oidcInfo,
        },
      };
    },
  })
);
