import {getSession} from '@auth0/nextjs-auth0';
import {GetServerSidePropsContext, GetServerSidePropsResult} from 'next';
import {Layout} from '../../components/layout';
import {ConcreteResource, User} from '../../database/models';
import {reportMetrics} from '../../lib/metrics';
import {getOrCreateUser} from '../../lib/get_user';
import {getResource} from '../../lib/queries';

interface ViewResourceProps {
  user: User | null;
  resource: ConcreteResource;
}

export default function ViewResource(props: ViewResourceProps) {
  const {user, resource} = props;
  let details = (
    <>
      <h1>
        {resource.resource_type}: {resource.resource_id}
      </h1>
      <p>Unsupported resource</p>
    </>
  );
  if (resource.resource_type === 'k8s') {
    details = (
      <>
        <h1>Kubernetes namespace: {resource.resource_id}</h1>
        {resource.cpu_quota ? (
          <p>
            CPU quota: <code>{resource.cpu_quota}</code>
          </p>
        ) : undefined}
        {resource.memory_quota ? (
          <p>
            Memory quota: <code>{resource.memory_quota}</code>
          </p>
        ) : undefined}
      </>
    );
  } else if (resource.resource_type === 'device') {
    details = (
      <>
        <h1>Device: {resource.resource_id}</h1>
        <p>MAC address: {resource.mac_address}</p>
      </>
    );
  }
  return (
    <Layout
      title={`${resource.resource_type}:${resource.resource_id}`}
      user={user}
    >
      {details}
    </Layout>
  );
}

export const getServerSideProps = reportMetrics(
  'resource',
  async (
    ctx: GetServerSidePropsContext
  ): Promise<GetServerSidePropsResult<ViewResourceProps>> => {
    // Parse the resource in the URL, 'type:id'
    const path: string = ctx.params!.path as string;
    const separator = path.indexOf(':');
    if (separator === -1) {
      return {notFound: true};
    }
    const resource = await getResource({
      resource_type: path.substring(0, separator),
      resource_id: path.substring(separator + 1),
    });
    if (resource === null) {
      return {notFound: true};
    }

    const session = getSession(ctx.req, ctx.res);
    let user = null;
    if (session) {
      user = await getOrCreateUser(session.user);
    }

    return {
      props: {
        user,
        resource,
      },
    };
  }
);
