import {getSession} from '@auth0/nextjs-auth0';
import {GetServerSidePropsContext, GetServerSidePropsResult} from 'next';
import {Resource, Space, Team, User} from '../../database/models';
import {reportMetrics} from '../../lib/metrics';
import {getOrCreateUser} from '../../lib/get_user';
import {
  getSpaceByName,
  getSubSpaces,
  getSpaceTeams,
  getSpaceOwners,
  getSpaceResources,
  getTeamAndMembers,
} from '../../lib/queries';
import {sortedByKey, sortByKeys, sortedByKeys} from '../../lib/utils';
import {Layout} from '../../components/layout';
import {
  LinkSpace,
  LinkSpaceAudit,
  LinkResource,
  LinkTeam,
} from '../../components/links';

interface ViewSpaceProps {
  user: User | null;
  space: Space;
  subpage: ViewSpaceIndex | ViewSpaceAudit | ViewSpaceTeam;
}

interface ViewSpaceIndex {
  which: 'index';
  spaces: Space[];
  teams: Team[];
  owners: Array<{user: User; inherited: boolean}>;
  resources: Array<Resource>;
}

interface ViewSpaceAudit {
  which: 'audit';
}

interface ViewSpaceTeam {
  which: 'team';
  team: Team;
  users: User[];
}

function Index(
  {space}: ViewSpaceProps,
  {spaces, teams, owners, resources}: ViewSpaceIndex
) {
  // Put inherited owners first and direct owners next, both sorted by name
  const sortedOwners = sortedByKeys(owners, ({user, inherited}) => [
    inherited ? 0 : 1,
    user.name,
  ]);
  // Sort teams by name
  const sortedTeams = sortedByKey(teams, team => team.name);
  // Get normal resources, the ones with non-empty type
  const sortedNormalResources = resources.filter(
    resource => resource.resource_type !== ''
  );
  sortByKeys(sortedNormalResources, ({resource_type, resource_id}) => [
    resource_type,
    resource_id,
  ]);
  // Get the special resources, the ones with empty type
  const sortedSpecialResources = resources.filter(
    resource => resource.resource_type === ''
  );
  sortByKeys(sortedSpecialResources, ({resource_type, resource_id}) => [
    resource_type,
    resource_id,
  ]);
  return (
    <>
      <p>Spaces:</p>
      <ul>
        {spaces.map(subspace => (
          <li key={subspace.space_id}>
            <LinkSpace name={space.fullname + '/' + subspace.name}>
              {subspace.name}
            </LinkSpace>
          </li>
        ))}
      </ul>
      <p>Owners:</p>
      <ul>
        {sortedOwners.map(({user, inherited}) => (
          <li key={user.user_id}>
            {user.name}
            {inherited ? ' (inherited)' : undefined}
          </li>
        ))}
      </ul>
      <p>Teams:</p>
      <ul>
        {sortedTeams.map(team => (
          <li key={team.team_id}>
            <LinkTeam spaceName={space.fullname} teamName={team.name}>
              {team.name}
            </LinkTeam>
          </li>
        ))}
      </ul>
      <p>Resources:</p>
      <ul>
        {sortedNormalResources.map(resource => (
          <li key={resource.resource_type + ':' + resource.resource_id}>
            <LinkResource res={resource}>
              {resource.resource_type}:{resource.resource_id}
            </LinkResource>
          </li>
        ))}
      </ul>
      <p>Creatable resources:</p>
      <ul>
        {sortedSpecialResources.map(resource => (
          <li key={resource.resource_id}>{resource.resource_id}</li>
        ))}
      </ul>
      <p>
        <LinkSpaceAudit name={space.fullname}>Audit</LinkSpaceAudit>
      </p>
    </>
  );
}

function Audit() {
  return <p>TODO: Space audit page</p>;
}

function Team({space}: ViewSpaceProps, {team, users}: ViewSpaceTeam) {
  const sortedUsers = sortedByKey(users, user => user.name);
  return (
    <>
      <p>Team name: {team.name}</p>
      <p>Members:</p>
      <ul>
        {sortedUsers.map(user => (
          <li key={user.user_id}>{user.name}</li>
        ))}
      </ul>
    </>
  );
}

export default function ViewSpace(props: ViewSpaceProps) {
  const {user, space, subpage} = props;
  let page;
  const displayName = space.fullname.replace('/', ' / ');
  if (subpage.which === 'audit') {
    page = Audit();
  } else if (subpage.which === 'team') {
    page = Team(props, subpage);
  } else {
    page = Index(props, subpage);
  }
  const spaceLinks = [];
  const components = space.fullname.split('/');
  for (let i = 0; i < components.length; ++i) {
    const otherSpace = components.slice(0, i + 1).join('/');
    if (i > 0) {
      spaceLinks.push(' / ');
    }
    spaceLinks.push(
      <LinkSpace key={i} name={otherSpace}>
        {components[i]}
      </LinkSpace>
    );
  }
  return (
    <Layout title={displayName} user={user}>
      <h1>{spaceLinks}</h1>
      {page}
    </Layout>
  );
}

export const getServerSideProps = reportMetrics(
  'space',
  async (
    ctx: GetServerSidePropsContext
  ): Promise<GetServerSidePropsResult<ViewSpaceProps>> => {
    // Parse the URL, which has the format 'full/space/name/-/some/route'
    const path: string[] = (ctx.params?.path as string[] | undefined) || [];
    const separator = path.indexOf('-');
    let spaceName: string[] = path;
    let route: string[] = [];
    if (separator !== -1) {
      spaceName = path.slice(0, separator);
      route = path.slice(separator + 1);
      if (route.length === 0) {
        // The URL looks like 'space/name/-'
        // Redirect to 'space/name'
        return {
          redirect: {destination: '/s/' + spaceName.join('/'), permanent: true},
        };
      }
    }
    const space = await getSpaceByName(spaceName);
    if (!space) {
      return {notFound: true};
    }

    const session = getSession(ctx.req, ctx.res);
    let user = null;
    if (session) {
      user = await getOrCreateUser(session.user);
    }

    // Pick the correct view based on URL
    let subpage;
    if (route.length === 0) {
      const [spaces, teams, owners, resources] = await Promise.all([
        getSubSpaces(space.space_id),
        getSpaceTeams(space.space_id),
        getSpaceOwners(space.space_id),
        getSpaceResources(space.space_id),
      ]);
      subpage = {
        which: 'index',
        spaces,
        teams,
        owners,
        resources,
      } as ViewSpaceIndex;
    } else if (route[0] === 'audit' && route.length === 1) {
      subpage = {which: 'audit'} as ViewSpaceAudit;
    } else if (route[0] === 'team' && route.length === 2) {
      const teamDetails = await getTeamAndMembers(space.space_id, route[1]);
      if (teamDetails === null) {
        return {notFound: true};
      }
      const {team, users} = teamDetails;
      subpage = {which: 'team', team, users} as ViewSpaceTeam;
    } else {
      return {notFound: true};
    }

    return {props: {user, space, subpage}};
  }
);
